/* SPDX-License-Identifier: GPL-2.0-only */
#ifndef LABWC_TABLET_TOOL_CONFIG_H
#define LABWC_TABLET_TOOL_CONFIG_H

#include "config/types.h"

enum lab_motion tablet_parse_motion(const char *name);

#endif /* LABWC_TABLET_TOOL_CONFIG_H */
